from dotenv import load_dotenv
import os

load_dotenv()
# Read .env file

DB_NAME = os.getenv('CONFIG_COBRA_DB_NAME')
DB_HOST = os.getenv('CONFIG_COBRA_DB_HOST')
DB_PORT = os.getenv('CONFIG_COBRA_DB_PORT')
DB_USER = os.getenv('CONFIG_COBRA_DB_USER')
DB_PASSWORD = os.getenv('CONFIG_COBRA_DB_PASSWORD')


import pytest
from unittest.mock import patch
from src.db.database_instance import SessionManager
from src.runner import UpdateRunner

@patch('src.db.database_instance.DatabaseInstance')
@patch('src.runner.update_runner.UpdateRunner._load_yaml')
@patch('src.runner.update_runner.UpdateRunner._check_table_exists')
def test_process_yaml(mock_check_table_exists, mock_load_yaml, mock_database_instance):
    mock_check_table_exists.return_value = True
    mock_load_yaml.return_value = {
        "deploy": [
            {
                "table": "test_table",
                "action": "insert",
                "columns": ["id"],
                "data": [
                    {
                        "id": 1
                    }
                ]
            }
        ]
    }

    update_runner = UpdateRunner(mock_database_instance)
    update_runner.process_yaml("test_file.yaml")

    assert mock_check_table_exists.called
    assert mock_load_yaml.called
    assert update_runner.yaml_data == {
        "deploy": [
            {
                "table": "test_table",
                "action": "insert",
                "columns": ["id"],
                "data": [
                    {
                        "id": 1
                    }
                ]
            }
        ]
    }
    assert update_runner.was_table_updated == True

@patch('src.runner.update_runner.UpdateRunner._load_yaml')
@patch('src.runner.update_runner.UpdateRunner._check_table_exists')
def test_process_yaml_invalid_table_name(mock_check_table_exists, mock_load_yaml):
    mock_check_table_exists.return_value = False
    mock_load_yaml.return_value = {
        "deploy": [
            {
                "table": "test_table",
                "columns": ["id"]
            }
        ]
    }

    update_runner = UpdateRunner(None)
    update_runner.process_yaml("test_file.yaml")

    assert mock_check_table_exists.called
    assert mock_load_yaml.called
    assert update_runner.yaml_data == {
        "deploy": [
            {
                "table": "test_table",
                "columns": ["id"]
            }
        ]
    }
    assert update_runner.was_table_updated == False


import logging
from sqlalchemy import create_engine, MetaData, inspect
from sqlalchemy.orm import sessionmaker, Session

logger = logging.getLogger("pipeline_data_deploy_logger")

class DatabaseInstance():

    def __init__(self, db_user, db_password, db_host, db_port, db_name):
        logger.info("Creating a database instance")
        self.db_user = db_user
        self.db_password = db_password
        self.db_host = db_host
        self.db_port = db_port
        self.db_name = db_name
        self._get_engine()

    def _get_engine(self):

        self.engine = create_engine(self.construct_uri())
        self.metadata = MetaData()
        self.Session = sessionmaker(bind=self.engine) 

    def construct_uri(self):

        return f"postgresql://{self.db_user}:{self.db_password}@{self.db_host}:{self.db_port}/{self.db_name}"
    
    def get_tables(self):
        insp = inspect(self.engine)
        return insp.get_table_names()
    
    def get_session(self) -> Session:
        return SessionManager(self.Session())
    
class SessionManager:
    def __init__(self, session):
        self.session = session

    def __enter__(self):
        return self.session

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.session.close()


from sqlalchemy.exc import IntegrityError
import yaml
import logging
from src.db.database_instance import DatabaseInstance
from sqlalchemy import text

logger = logging.getLogger("pipeline_data_deploy_logger")

def validate_local_fields(func):
    """
    Decorator to check data is as described in yaml file
    """
    def decorator(self, *args, **kwargs):
        
        if not all(key in self.columns for key in kwargs['data'].keys()):
            logger.error(f"Data does not match yaml file, expected fields: {self.columns}")
            return None
        
        if not all(key in kwargs['data'].keys() for key in self.columns):
            logger.error(f"Data does not match yaml file, expected fields: {self.columns}")
            return None

        return func(self, *args, **kwargs)
    
    return decorator
    
class UpdateRunner():

    def __init__(self, db: DatabaseInstance):
        self.db = db
        self.yaml_data = None
        self.was_table_updated = False
        self.columns = None

    def process_yaml(self, yaml_file):

        self.yaml_data = self._load_yaml(yaml_file)

        for deploy in self.yaml_data["deploy"]:

            self.columns = deploy["columns"]

            if self._check_table_exists(deploy["table"]):
                logger.info("Table exists, updating table")

                if deploy["action"] == "insert":
                    for data in deploy["data"]:
                        self._is_insert(data=data, table=deploy["table"])
                elif deploy["action"] == "update":
                    for data in deploy["data"]:
                        self._is_update(data=data, table=deploy["table"], pk_column=deploy["pk"])
                elif deploy["action"] == "update_or_insert":
                    for data in deploy["data"]:
                        self._is_update_or_insert(data=data, table=deploy["table"], pk_column=deploy["pk"])
                elif deploy["action"] == "delete":
                    for data in deploy["data"]:
                        self._is_delete(data=data, table=deploy["table"], pk_column=deploy["pk"])

                self.was_table_updated = True
            else:
                logger.warn("Table does not exist, will not update table")

    @validate_local_fields
    def _is_insert(self, data:dict, table:str):

        with self.db.get_session() as session:
            try:
                query = text(f"INSERT INTO {table} ({','.join(data.keys())}) VALUES ({','.join(':' + key for key in data.keys())})")
                session.execute(query, params=data)
                session.commit()
            except IntegrityError as e:
                logger.error(f"Error inserting data: {e}")
                session.rollback()

    @validate_local_fields
    def _is_update(self, data:dict, table:str, pk_column:str="id"):
            
        with self.db.get_session() as session:
            try:
                logger.info(f"Updating data in table {table} where {pk_column} = {data[pk_column]}")
                query = text(f"UPDATE {table} SET {','.join(f'{key} = :{key}' for key in data.keys())} WHERE {pk_column} = :{pk_column}")
                session.execute(query, params=data)
                session.commit()
            except IntegrityError as e:
                logger.error(f"Error updating data: {e}")
                session.rollback()

    @validate_local_fields
    def _is_update_or_insert(self, data:dict, table:str, pk_column:str="id"):
            
            with self.db.get_session() as session:
                try:
                    logger.info(f"Updating or inserting data into table {table} where {pk_column} = {data[pk_column]}")
                    query = text(f"INSERT INTO {table} ({','.join(data.keys())}) VALUES ({','.join(':' + key for key in data.keys())}) ON CONFLICT ({pk_column}) DO UPDATE SET {','.join(f'{key} = :{key}' for key in data.keys())}")
                    session.execute(query, params=data)
                    session.commit()
                except IntegrityError as e:
                    logger.error(f"Error updating or inserting data: {e}")
                    session.rollback()

    @validate_local_fields
    def _is_delete(self, data:dict, table:str, pk_column:str="id"):
            
        with self.db.get_session() as session:
            try:
                logger.info(f"Deleting data from table {table} where {pk_column} = {data[pk_column]}")
                query = text(f"DELETE FROM {table} WHERE {pk_column} = :{pk_column}")
                session.execute(query, params=data)
                session.commit()
            except IntegrityError as e:
                logger.error(f"Error deleting data: {e}")
                session.rollback()

    def _check_table_exists(self, table_name:str):

        if table_name in self.db.get_tables():
            return True
        
        return False

    def _load_yaml(self, yaml_file):
        with open(yaml_file, 'r') as stream:
            try:
                return yaml.safe_load(stream)
            except yaml.YAMLError as exc:
                print(exc)
                return None


from setuptools import setup, find_packages

# Parse requirements.txt
with open('requirements.txt') as f:
    requirements = f.read().splitlines()

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name='config-cobra',
    version='0.1.6',
    description='A simple yaml to database updater',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/JTSG1/configCobra',
    license='MIT',
    packages=find_packages(),
    include_package_data=True,
    install_requires=requirements,
    entry_points={
        'console_scripts': [
            'config-cobra=src.app:main',
        ],
    },
)


import argparse
import logging
from src.db import DatabaseInstance
from src.settings import settings
from src.runner import UpdateRunner

logger = logging.getLogger("pipeline_data_deploy_logger")
logger.setLevel(logging.INFO)
handler = logging.StreamHandler()
logger.addHandler(handler)
formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
handler.setFormatter(formatter)

def main():

    logger.info("Starting the update process")

    parser = argparse.ArgumentParser(description="Update the database with from yaml file")
    parser.add_argument('--yaml_file', '-y', type=str, help="The yaml file to update the database with", required=True)

    args = parser.parse_args()

    db = DatabaseInstance(
        db_user=settings.DB_USER,
        db_password=settings.DB_PASSWORD,
        db_host=settings.DB_HOST,
        db_port=settings.DB_PORT,
        db_name=settings.DB_NAME
    )

    update_runner = UpdateRunner(db)

    if "yaml_file" in args:
        update_runner.process_yaml(args.yaml_file)


    print(db.get_tables())

if __name__ == '__main__':
    main()